(require 'icalendar)

(defun gnokii-diary-fetch ()
  "Fetch diary entries from phone."
  (interactive)
  ;; vcalendar files have CRLF
  (let ((coding-system-for-read 'utf-8-dos))
    (set-process-sentinel
     (start-process "gnokii-diary" 
		    (generate-new-buffer-name "gnokii-diary")
		    "gnokii" "--getcalendarnote" "1" "end" "-v")
     'gnokii-diary-fetch-done)))

(defun gnokii-diary-send ()
  "Send diary entries to phone."
  (interactive)
  ;; vcalendar files have CRLF
  (let ((coding-system-for-read 'utf-8-dos))
    (set-process-sentinel
     (start-process "gnokii-diary" 
		    (generate-new-buffer-name "gnokii-diary")
		    "gnokii" "--getcalendarnote" "1" "end" "-v")
     'gnokii-diary-do-send)))

(defun gnokii-diary-fetch-done (process string)
  (let ((entries
	 (with-current-buffer (process-buffer process)
	   (gnokii-diary-parse-vcalendar))))
    (dolist (entry entries)
      (destructuring-bind (year month day summary) entry
	;; Only use the entry if it happens today or later.
	(when (time-less-p (current-time)
			   (encode-time 0 0 24 day month year))
	  (let ((diary-entry 
		 (format "%s %s"
			 (calendar-date-string
			  (list month day year) t t)
			 summary)))
	    (with-temp-buffer
	      (insert-file-contents (substitute-in-file-name diary-file))
	      (goto-char (point-min))
	      (unless (search-forward diary-entry nil t)
		(when (y-or-n-p (format "Insert \"%s\" into diary? " diary-entry))
		  (make-diary-entry diary-entry))))))))
    (message "%d calendar notes read from phone" (length entries))))

(defun gnokii-diary-parse-vcalendar ()
  "Parse all vcalendar events in current buffer.
Return a list of entries, each being (YEAR MONTH DAY SUMMARY)."
  (goto-char (point-min))
  (let (entries)
    (while (search-forward "BEGIN:VEVENT" nil t)
      (let ((bound (save-excursion 
		     (search-forward "END:VEVENT")
		     (point))))
	(let* ((summary (save-excursion
			  (search-forward-regexp "^SUMMARY:\\(\\(\n \\|[^\n]\\)+\\)$" bound)
			  (gnokii-decode-ical-string (match-string 1))))
	       (date-string (save-excursion
			      (and (search-forward-regexp "^DTSTART:\\([0-9]*T[0-9]*\\)" bound t)
				   (match-string 1)))))
	  (when date-string
	    (let (
	       ;; date is in format "YYYYMMDDTHHMMSS".  If AALARM is
	       ;; present, it is the same as DTSTART, just meaning
	       ;; that an alarm was set.
	       (year (string-to-number (substring date-string 0 4)))
	       (month (string-to-number (substring date-string 4 6)))
	       (day (string-to-number (substring date-string 6 8))))
	      (push (list year month day (gnokii-simplify-whitespace summary)) entries))))))
    (nreverse entries)))

(defvar gnokii-diary-widget nil)

(defun gnokii-diary-do-send (process string)
  (let* (;; (phone-entries (with-current-buffer (process-buffer process)
;; 			  (gnokii-diary-parse-vcalendar)))
	 (calendar-today
	  (destructuring-bind (seconds minutes hour day month year down dst zone)
	      (decode-time)
	    (list month day year)))
	 (diary-entries (diary-list-entries calendar-today 90 t)))
    ;;(message "phone: %S\ndiary: %S" phone-entries diary-entries)
    ;;(unless (y-or-n-p (format "Read %d entries from phone and %d entries from diary.  Proceed? "
	;;		      (length phone-entries) (length diary-entries)))
      ;;(error "Cancelled"))
    (with-current-buffer (get-buffer-create "gnokii-calendar")
      (erase-buffer)
      (fundamental-mode)
      (set (make-local-variable
	    'gnokii-diary-widget)
	   (widget-create
	    '(repeat
	      :tag "Calendar entries"
	      (list :tag "Entry"
		    (list :tag "Date"
			  (integer :tag "Month")
			  (integer :tag "Day")
			  (integer :tag "Year"))
		    (text :tag "Text")
		    (choice
		     (const :tag "No alarm" nil)
		     (list :tag "Alarm"
			   (integer :tag "Hour")
			   (integer :tag "Minute")))))
	    :value
	    (mapcar
	     (lambda (entry)
	       (destructuring-bind ((month day year) summary &rest ignored) entry
		 (list (list month day year) (gnokii-simplify-whitespace summary) nil)))
	     diary-entries)))
      (widget-insert "\n")
      (widget-create 'push-button :notify 'gnokii-diary-really-do-send "Send to phone")

      (widget-setup)
      (widget-minor-mode 1)
      (switch-to-buffer (current-buffer))
      (goto-char (point-min)))))

(defun gnokii-diary-really-do-send (&rest ignore)
  (let ((file-name (make-temp-file "gnokii-diary"))
	(entries (widget-value gnokii-diary-widget)))
    (with-temp-buffer
      (insert "BEGIN:VCALENDAR
VERSION:1.0
")
      (dolist (entry entries)
	(destructuring-bind ((month day year) summary (&optional hour minute)) entry
	  (setq summary (gnokii-simplify-whitespace summary))
	  (unless (or
		   ;;(member (list year month day summary) phone-entries)
		   (time-less-p (encode-time 0 0 24 day month year)
				(current-time)))
	    (insert "BEGIN:VEVENT
CATEGORIES:REMINDER
SUMMARY:"
		    (gnokii-encode-ical-string summary) "\n")
	    (let* ((alarm (or hour minute))
		   ;; if there is no alarm, fake the event starting at 09:00
		   (hour (or hour 9))
		   (minute (or hour 0))
		   (second 0)
		   (time-string (format "%04d%02d%02dT%02d%02d%02d" year month day hour minute second)))
	      (insert "DTSTART:" time-string "\n")
	      (when alarm
		(insert "AALARM:" time-string "\n")))
	    (insert "END:VEVENT\n"))))
      (insert "END:VCALENDAR\n")
      (write-region (point-min) (point-max) file-name)
      (message "Sending calendar entries...")
      (unless (equal (call-process "gnokii" nil "*gnokii-diary*" nil "--writecalendarnote" file-name "1" "end") 0)
	(error "Sending calendar entries failed"))
      (message "Sending calendar entries...done"))))

(defun gnokii-simplify-whitespace (s)
  (replace-regexp-in-string "^[ \t]+\\|[ \t]+$" "" s))

(defun gnokii-encode-ical-string (s)
  (dolist (replace '(("\\\\" . "\\\\")
		     (";" . "\\\\;")
		     ("," . "\\\\,")
		     ("\n" . "\\\\n"))
		   s)
    (setf s (replace-regexp-in-string (car replace) (cdr replace) s))))

(defun gnokii-decode-ical-string (s)
  (dolist (replace '(("\n " . "")
		     ("\\\\n" . "\n")
		     ("\\\\N" . "\n")
		     ("\\\\;" . ";")
		     ("\\\\," . ",")
		     ("\\\\\\\\" . "\\\\"))
		   s)
    (setf s (replace-regexp-in-string (car replace) (cdr replace) s))))
